/************************************************************************/
/* sdlsysc.c: SDL library wrapper for image display in SystemC models	*/
/*									*/
/* 03/07/16 RD	added XInitThreads() call for MT-safe X graphics	*/
/* 03/26/15 GL	original version (based on sdlspecc.c)			*/
/************************************************************************/ 

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <SDL.h>
#include <X11/Xlib.h>

#include "sdlsysc.h"

/************************************************************************/ 

SDL_Surface	*_SDL_SysC_Screen = NULL;	/* global handle to the window */

static SDL_sem	*_SDL_SysC_ScreenLock = NULL;	/* lock for safe rendering */


/************************************************************************/ 

void _SDL_SysC_InitDisplay(unsigned int w, unsigned int h, const char *title)
{
    Uint32	vflags;
    int		bpp;
    const SDL_VideoInfo	*info;

    /* initialize X graphics for multi-threading (RD, 03/07/16) */
    if (!XInitThreads())
    {   fprintf(stderr, "SDL_SYSC: Failed to initialize X for multi-threading\n");
	exit(100);
    }
    /* initialize SDL's video system and check for errors */
    if (SDL_Init(SDL_INIT_VIDEO) != 0)
    {   fprintf(stderr, "SDL_SYSC: Failed to initialize video: %s\n", SDL_GetError());
	exit(101);
    }
    /* make sure SDL_Quit gets called when the program exits */
    atexit(SDL_Quit);

    /* ensure we get a window with true-color mode (8 bits each for R,G,B) */
    if (!(info = SDL_GetVideoInfo()))
    {   fprintf(stderr, "SDL_SYSC: Video query failed: %s\n", SDL_GetError());
	exit(102);
    }
    assert(info->vfmt);
    bpp = info->vfmt->BitsPerPixel;
    if (  (bpp != 32)
	||(info->vfmt->Rloss != 0)
	||(info->vfmt->Gloss != 0)
	||(info->vfmt->Bloss != 0)
	||(info->vfmt->Rshift != 16)
	||(info->vfmt->Gshift != 8)
	||(info->vfmt->Bshift != 0)
	||(info->vfmt->Rmask != 255<<16)
	||(info->vfmt->Gmask != 255<<8)
	||(info->vfmt->Bmask != 255))
    {   fprintf(stderr, "SDL_SYSC: Unsupported video format (%d,%d,%d,%d),\n"
			"SDL_SYSC: need true-color mode (32,16,8,0)!\n",
		bpp, info->vfmt->Rshift, info->vfmt->Gshift, info->vfmt->Bshift);
	exit(103);
    }
    if (info->hw_available)
    {   vflags = SDL_HWSURFACE;
    }
    else	
    {   vflags = SDL_SWSURFACE;
    }
    if (title)
    {   SDL_WM_SetCaption(title, title);
    }
    if (!(_SDL_SysC_Screen = SDL_SetVideoMode(w, h, bpp, vflags)))
    {   fprintf(stderr, "SDL_SYSC: Set video mode failed: %s\n", SDL_GetError());
	exit(104);
    }
    assert(_SDL_SysC_Screen->format);
    bpp = _SDL_SysC_Screen->format->BitsPerPixel;
    if (  (bpp != 32)
	||(_SDL_SysC_Screen->format->Rloss != 0)
	||(_SDL_SysC_Screen->format->Gloss != 0)
	||(_SDL_SysC_Screen->format->Bloss != 0)
	||(_SDL_SysC_Screen->format->Rshift != 16)
	||(_SDL_SysC_Screen->format->Gshift != 8)
	||(_SDL_SysC_Screen->format->Bshift != 0)
	||(_SDL_SysC_Screen->format->Rmask != 255<<16)
	||(_SDL_SysC_Screen->format->Gmask != 255<<8)
	||(_SDL_SysC_Screen->format->Bmask != 255))
    {   fprintf(stderr, "SDL_SYSC: Got unsupported video format (%d,%d,%d,%d),\n"
			"SDL_SYSC: need true-color mode (32,16,8,0)!\n",
		bpp, _SDL_SysC_Screen->format->Rshift,
		_SDL_SysC_Screen->format->Gshift,
		_SDL_SysC_Screen->format->Bshift);
	exit(105);
    }
    /* ensure that we don't need to lock the screen for drawing pixels */
    if (SDL_MUSTLOCK(_SDL_SysC_Screen))
    {   fprintf(stderr, "SDL_SYSC: Must lock screen, not supported!\n");
	exit(106);
    }
    /* finally create a semaphore for safe concurrent updates */
    if (!(_SDL_SysC_ScreenLock = SDL_CreateSemaphore(1)))
    {   fprintf(stderr, "SDL_SYSC: Create semaphore failed!\n");
	exit(107);
    }
    /* all done! */
}

void _SDL_SysC_DrawPixel(unsigned char r, unsigned char g, unsigned char b,
				unsigned int x, unsigned int y)
{
    Uint32	*raw_pixels;
    Uint32	pixel_color;
    int		offset;

    assert(_SDL_SysC_Screen);
    assert(x < _SDL_SysC_Screen->w);
    assert(y < _SDL_SysC_Screen->h);

    raw_pixels = (Uint32*)_SDL_SysC_Screen->pixels;
    pixel_color = (r << 16) | (g << 8) | b;
    offset = (_SDL_SysC_Screen->pitch/sizeof(Uint32) * y + x);
    raw_pixels[offset] = pixel_color;
}

void _SDL_SysC_UpdateRect(unsigned int x, unsigned int y,
				unsigned int w, unsigned int h)
{
    assert(_SDL_SysC_Screen);
    assert(_SDL_SysC_ScreenLock);
    assert(x < _SDL_SysC_Screen->w);
    assert(y < _SDL_SysC_Screen->h);
    assert(w <= _SDL_SysC_Screen->w);
    assert(h <= _SDL_SysC_Screen->h);

    SDL_SemWait(_SDL_SysC_ScreenLock);
    SDL_UpdateRect(_SDL_SysC_Screen, x, y, w, h);
    SDL_SemPost(_SDL_SysC_ScreenLock);
}

void _SDL_SysC_ClearDisplay(void)
{
    assert(_SDL_SysC_Screen);
    assert(_SDL_SysC_ScreenLock);

    SDL_SemWait(_SDL_SysC_ScreenLock);
    memset(_SDL_SysC_Screen->pixels, 0,
		_SDL_SysC_Screen->w * _SDL_SysC_Screen->h * sizeof(Uint32));
    SDL_UpdateRect(_SDL_SysC_Screen, 0, 0, 0, 0);
    SDL_SemPost(_SDL_SysC_ScreenLock);
}

void _SDL_SysC_FillDisplay(unsigned char r, unsigned char g, unsigned char b)
{
    assert(_SDL_SysC_Screen);
    assert(_SDL_SysC_ScreenLock);

    SDL_SemWait(_SDL_SysC_ScreenLock);
    memset(_SDL_SysC_Screen->pixels, (r<<16)|(g<<8)|b,
		_SDL_SysC_Screen->w * _SDL_SysC_Screen->h * sizeof(Uint32));
    SDL_UpdateRect(_SDL_SysC_Screen, 0, 0, 0, 0);
    SDL_SemPost(_SDL_SysC_ScreenLock);
}

int  _SDL_SysC_DisplayClosed(void)
{
    return SDL_QuitRequested();
}

/* EOF */
