#include "class.h"

#include "design.h"
#include "event.h"
#include "event_and_list.h"
#include "event_or_list.h"
#include "function.h"
#include "hierarchical_channel.h"
#include "hierarchical_channel_instance.h"
#include "inoutport.h"
#include "inport.h"
#include "module.h"
#include "module_instance.h"
#include "outport.h"
#include "port.h"
#include "port_mapping.h" // needed for risc::get_mapped_symbol(...)
#include "primitive_channel_instance.h"
#include "variable.h"

#include "../tools/class_is_derived_from.h"
#include "../tools/misc.h"

risc::Class::Class(SgClassDefinition *ast_node, Type type):
  Definition(ast_node, type)
{
  initialize_variables();
}

risc::Class::~Class()
{
  for(VariableVectorIter
      iter  = variables_.begin();
      iter != variables_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(PrimitiveChannelInstanceVectorIter
      iter  = sub_prim_channel_.begin();
      iter != sub_prim_channel_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(HierarchicalChannelInstanceVectorIter
      iter  = sub_hier_channel_.begin();
      iter != sub_hier_channel_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(ModuleInstanceVectorIter
      iter  = sub_modules_.begin();
      iter != sub_modules_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(FunctionVectorIter
      iter  = functions_.begin();
      iter != functions_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(EventVectorIter
      iter  = events_.begin();
      iter != events_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(EventAndListVectorIter
      iter  = event_and_lists_.begin();
      iter != event_and_lists_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(EventOrListVectorIter
      iter  = event_or_lists_.begin();
      iter != event_or_lists_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(InPortVectorIter
      iter  = inports_.begin();
      iter != inports_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(InOutPortVectorIter
      iter  = inoutports_.begin();
      iter != inoutports_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(OutPortVectorIter
      iter  = outports_.begin();
      iter != outports_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }
}

void
risc::Class::initialize_functions()
{
  // get ctor of the class (module)
  SgMemberFunctionDeclaration *ctor =
    SageInterface::getDefaultConstructor(get_ast_node()->get_declaration());

  // get dtor of the class (module)
  SgMemberFunctionDeclaration *dtor =
    SageInterface::getDefaultDestructor(get_ast_node()->get_declaration());

  // Find all functions in the class defintion
  std::vector<SgNode*> function_definitions =
    NodeQuery::querySubTree(get_ast_node(), V_SgFunctionDefinition);

  // The constructor is in the function_definitions container because rose
  // treats a constructor as a special kind of a function. So, we have to
  // search for the ctor and delete it.
  function_definitions.erase(
    std::remove(function_definitions.begin(), function_definitions.end(),
             ctor->get_definition()),
    function_definitions.end());

  // same for the destructor
  if(dtor) {
    function_definitions.erase(
      std::remove(function_definitions.begin(), function_definitions.end(),
               dtor->get_definition()),
      function_definitions.end());
  }
}

SgClassDefinition*
risc::Class::get_ast_node()
{
  return isSgClassDefinition(ast_node_);
}

std::string
risc::Class::get_ast_type_name()
{
  return get_ast_node()->get_qualified_name().getString();
}

std::string risc::Class::get_name()
{
  return SageInterface::get_name(get_ast_node());
}

void
risc::Class::initialize_variables()
{
  // This loop iterates all members of the class. A member can be a
  // constructor, typedef, function, and so on. For this case we have to
  // distinguish the different types
  for(SgDeclarationStatementPtrList::iterator
      iter  = get_ast_node()->get_members().begin();
      iter != get_ast_node()->get_members().end();
      iter++) {

    SgVariableDeclaration *vd = isSgVariableDeclaration(*iter);
    bool is_array = false;
    bool is_pointer = false;

    // We are only interested in variable declarations
    if(!vd) {

      continue;
    }

    SgVariableDefinition *var_def = vd->get_definition();

    SgType* type = var_def->get_type();

    if(isSgReferenceType(type)) {

      type = isSgReferenceType(type)->get_base_type();
    }

    if(isSgTypedefType(type)) {

      type = isSgTypedefType(type)->get_base_type();
    }

    // perhaps we have an array of pointers.
    // We are analyzing this case here.
    if(isSgArrayType(type)) {

    }

    if(SageInterface::isPointerType(type)) {

    }

    // We check if it is a SystemC element
    //
    // A sc_module instance has two significant attributes:
    //  a) the instance is a data structure (class)
    //  b) the class is derived from sc_module
    //
    // A port
    //  a) derived from sc_port
    if(risc::tools::is_data_structure(type)) {

      SgClassType *ct = isSgClassType(type);

      SgClassDeclaration *declaration =
        isSgClassDeclaration(ct->get_declaration());

      if(declaration->isForward()) {
        declaration =
          isSgClassDeclaration(declaration->get_definingDeclaration());
      }

      SgClassDefinition *definition = declaration->get_definition();

      //Okay, at this point we know it is a C++ data structure.

      // case of a channel
      if(risc::tools::is_derived_from_sc_channel(definition) ) {

        HierarchicalChannel *hc =
          GlobalDesign::design_static->find_channel_definition(definition);

        if(hc == NULL) {
          get_design()->add_hierarchical_channel_to_defintion_list(definition);
          hc = GlobalDesign::design_static->find_channel_definition(definition);
        }

        sub_hier_channel_.push_back(
            new HierarchicalChannelInstance(var_def, hc));
        continue;
      }

      // case of a primitive channel
      if(risc::tools::is_derived_from_sc_prim_channel(definition)) {

        PrimitiveChannel *pc =
          GlobalDesign::design_static->find_prim_channel_definition(definition);

        // Triggers if no channel definition is available
        assert(pc);

        sub_prim_channel_.push_back(new PrimitiveChannelInstance(var_def, pc));
        continue;
      }

      //case of a module
      if(risc::tools::is_derived_from_sc_module(definition) ) {

        // At this point, we have the SgClassDefinition of the corresponding
        // class member variable in the module. In the next step, we have to
        // create a Module instance and insert into the design

        Module *md =
          GlobalDesign::design_static->find_module_definition(definition);

        if(md == NULL) {
          get_design()->add_module_to_definition_list(definition);
          md = GlobalDesign::design_static->find_module_definition(definition);
        }

        sub_modules_.push_back(new ModuleInstance(var_def, md));
        continue;
      }

      // case of a port
      if(risc::tools::is_derived_from_sc_port_base(definition) ) {

        SgTemplateInstantiationDefn* templ_def =
          isSgTemplateInstantiationDefn(definition);
        if(templ_def) {

          // case of an in_port
          if(risc::tools::is_sc_in_port(templ_def)){

            inports_.push_back(new InPort(var_def, this));
            continue;
          }

          // case of an out_port
          if(risc::tools::is_sc_out_port(templ_def)){

            outports_.push_back(new OutPort(var_def, this));
            continue;
          }

          // case of an inout_port
          if(risc::tools::is_sc_inout_port(templ_def)){

            inoutports_.push_back(new InOutPort(var_def, this));
            continue;
          }
        }

        SgClassDefinition *class_def = isSgClassDefinition(definition);
        if(class_def) {

          // case of an sc_in_clk
          if(risc::tools::is_sc_in_clk(class_def)){

            inports_.push_back(new InPort(var_def, this));
            continue;
          }

          // case of an sc_out_clk
          if(risc::tools::is_sc_out_clk(class_def)){

            outports_.push_back(new OutPort(var_def, this));
            continue;
          }

          // case of an sc_inout_clk
          if(risc::tools::is_sc_inout_clk(class_def)){

            inoutports_.push_back(new InOutPort(var_def, this));
            continue;
          }
        }

        // in any other case we can not define it. So it must be an inout port
        inoutports_.push_back(new InOutPort(var_def, this));
        continue;
      }

      // case of an event
      if(risc::tools::is_sc_event(definition)) {

        events_.push_back(new Event(var_def));
        continue;
      }

      // case of an event and list
      if(risc::tools::is_sc_event_and_list(definition)) {

        event_and_lists_.push_back(new EventAndList(var_def));
        continue;
      }

      // case of an event or list
      if(risc::tools::is_sc_event_or_list(definition)) {

        event_or_lists_.push_back(new EventOrList(var_def));
        continue;
      }
    }

    // We check if it is an int, double, float, unsigned, ...
    if(risc::tools::is_builtin_type(type) ||
       risc::tools::is_data_structure(type)) {

      variables_.push_back(new Variable(var_def));
      continue;
    }

    if(SageInterface::isReferenceType(var_def->get_type())) {

      variables_.push_back(new Variable(var_def));
    }
  }
}

risc::Port*
risc::Class::find_port(SgVariableDefinition* var_def)
{
  for(InPortVectorConstIter
      iter  = inports_.begin();
      iter != inports_.end();
      iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  for(InOutPortVectorConstIter
      iter  = inoutports_.begin();
      iter != inoutports_.end();
      iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  for(OutPortVectorConstIter
      iter  = outports_.begin();
      iter != outports_.end();
      iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

risc::Event*
risc::Class::find_event(SgVariableDefinition* var_def)
{
  for(EventVectorConstIter
      iter  = events_.begin();
      iter != events_.end();
      iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

risc::EventAndList*
risc::Class::find_event_and_list(SgVariableDefinition* var_def)
{
  for(EventAndListVectorConstIter
      iter  = event_and_lists_.begin();
      iter != event_and_lists_.end();
      iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

risc::EventOrList*
risc::Class::find_event_or_list(SgVariableDefinition* var_def)
{
  for(EventOrListVectorConstIter
      iter  = event_or_lists_.begin();
      iter != event_or_lists_.end();
      iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

risc::Variable*
risc::Class::find_variable(SgVariableDefinition* var_def)
{
  for(VariableVectorConstIter
      iter  = variables_.begin();
      iter != variables_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

SgFunctionDefinition*
risc::Class::get_before_end_of_elaboration_definition()
{

  for(SgDeclarationStatementPtrList::iterator
      iter  = get_ast_node()->get_members().begin();
      iter != get_ast_node()->get_members().end();
      iter++) {

    if(SageInterface::get_name(*iter) == "before_end_of_elaboration") {
      SgFunctionDeclaration *func_decl = isSgFunctionDeclaration(*iter);
      func_decl = isSgFunctionDeclaration(func_decl->get_definingDeclaration());

      return func_decl->get_definition();
    }
  }

  return NULL;
}

risc::PrimitiveChannelInstance*
risc::Class::find_prim_channel(SgVariableDefinition* var_def)
{
  for(PrimitiveChannelInstanceVectorConstIter
      iter  = sub_prim_channel_.begin();
      iter != sub_prim_channel_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  return NULL;
}

risc::HierarchicalChannelInstance*
risc::Class::find_hier_channel(SgVariableDefinition* var_def)
{
  for(HierarchicalChannelInstanceVectorConstIter
      iter  = sub_hier_channel_.begin();
      iter != sub_hier_channel_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  return NULL;
}

risc::ModuleInstance*
risc::Class::find_module(SgVariableDefinition* var_def)
{
  for(ModuleInstanceVectorConstIter
      iter  = sub_modules_.begin();
      iter != sub_modules_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  return NULL;
}

risc::Instance*
risc::Class::find_instance(SgVariableDefinition* var_def)
{
  Instance *result = NULL;
  result = find_module(var_def);
  if(result) {
    return result;
  }

  result = find_hier_channel(var_def);
  if(result) {
    return result;
  }

  result = find_prim_channel(var_def);
  return result;
}

void risc::Class::initialize_mapping()
{
  // Strategy
  //
  // Here, we iterate over each module/hierarchical channel definition.
  // For a definition we analyze how the sub module instances are connected to
  // the hosting module.
  // The module instance has a map <SubModuleInstancePort, Host Port/Channel>

  // We only need to cast to Module, hierarichal channel is only derived
  Module *parent_module = dynamic_cast<Module*>(this);

  for(ModuleInstanceVectorConstIter
      module_instance_iter  = sub_modules_.begin();
      module_instance_iter != sub_modules_.end();
      module_instance_iter++) {

    Module *child_module = (*module_instance_iter)->get_definition();

    for(InPortVectorConstIter
        port_iter  = child_module->inports_.begin();
        port_iter != child_module->inports_.end();
        port_iter++) {

      analyze_port_mapping(parent_module, *module_instance_iter, *port_iter);
    }

    for(InOutPortVectorConstIter
        port_iter  = child_module->inoutports_.begin();
        port_iter != child_module->inoutports_.end();
        port_iter++) {

      analyze_port_mapping(parent_module, *module_instance_iter, *port_iter);
    }

    for(OutPortVectorConstIter
        port_iter  = child_module->outports_.begin();
        port_iter != child_module->outports_.end();
        port_iter++) {

      analyze_port_mapping(parent_module, *module_instance_iter, *port_iter);
    }
  }

  for(HierarchicalChannelInstanceVectorConstIter
      hier_channel_instance_iter  = sub_hier_channel_.begin();
      hier_channel_instance_iter != sub_hier_channel_.end();
      hier_channel_instance_iter++) {

    HierarchicalChannel *child_channel
      = (*hier_channel_instance_iter)->get_definition();

    for(InPortVectorConstIter
        port_iter  = child_channel->inports_.begin();
        port_iter != child_channel->inports_.end();
        port_iter++) {

      analyze_port_mapping(
          parent_module,
          *hier_channel_instance_iter,
          *port_iter);
    }

    for(InOutPortVectorConstIter
        port_iter  = child_channel->inoutports_.begin();
        port_iter != child_channel->inoutports_.end();
        port_iter++) {

      analyze_port_mapping(
          parent_module,
          *hier_channel_instance_iter,
          *port_iter);
    }

    for(OutPortVectorConstIter
        port_iter  = child_channel->outports_.begin();
        port_iter != child_channel->outports_.end();
        port_iter++) {

      analyze_port_mapping(
          parent_module,
          *hier_channel_instance_iter,
          *port_iter);
    }
  }
}

void risc::Class::analyze_port_mapping(
    Module *parent_module,
    ModuleInstance *sub_module_instance,
    Port *port)
{
  SgVariableSymbol *mapping
    = risc::get_mapped_symbol(port->get_symbol(),
        dynamic_cast<Module*>(this),
        sub_module_instance->get_ast_node());

  if(mapping == NULL) {
    //if(port->bounded) return;
    std::cout << "Error: Unsolved mapping."
              << std::endl;
    std::cout << "       Port "
              << port->get_name()
              << " of instance "
              << sub_module_instance->get_name()
              << " in "
              << parent_module->get_name()
              << " is not bound."
              << std::endl;
    exit(1);
  }

  SgVariableDefinition *parent_mapping
    = isSgVariableDefinition(mapping->get_declaration()->get_definition());

  Object *obj = parent_module->find_port(parent_mapping);

  if(obj == NULL) {
    obj = parent_module->find_prim_channel(parent_mapping);
  }
  if(obj == NULL) {
    obj = parent_module->find_hier_channel(parent_mapping);
  }

  if(obj == NULL) {
    std::cout << "Error: Unsolved mapping; Port "
              << port->get_name()
              << " of instance "
              << sub_module_instance->get_name()
              << " in "
              << parent_module->get_name()
              << " is not bound." << std::endl;
    exit(1);
  }
  
  //std::cout << "\nclass.cpp 681 port " << port->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString() << std::endl;
  //std::cout << "class.cpp 682 obj " << obj->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString() << std::endl; 
  sub_module_instance->add_mapping(port, obj);
}

bool risc::Class::has_module_pointers()
{
  SgDeclarationStatementPtrList &temp = get_ast_node()->get_members();

  for(std::vector<SgDeclarationStatement*>::iterator
      iter_decl  = temp.begin();
      iter_decl != temp.end();
      iter_decl++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter_decl);

    if (var_decl) {

      SgPointerType *type_pointer
        = isSgPointerType(
            var_decl->get_definition()->get_vardefn()->get_type());

      if (type_pointer) {


        // Simple pointer
        SgClassType *type = isSgClassType(type_pointer->get_base_type());

        if (type) {

          SgClassDeclaration *class_decl
            = isSgClassDeclaration(type->get_declaration());

          class_decl
            = isSgClassDeclaration(class_decl->get_definingDeclaration());

          Module *module
            = risc::get_design()->find_module_definition(
                                    class_decl->get_definition());

          if (module) {
            return true;
          }
        }


        // pointer to a pointer
        type_pointer = isSgPointerType(type_pointer->get_base_type());

        if (type_pointer) {

          SgClassType *class_type
            = isSgClassType(type_pointer->get_base_type());

          if (class_type) {

            SgClassDeclaration *class_decl
              = isSgClassDeclaration(class_type->get_declaration());

            class_decl
              = isSgClassDeclaration(class_decl->get_definingDeclaration());

            Module *module
              = risc::get_design()->find_module_definition(
                                      class_decl->get_definition());

            if (module) {
              return true;
            }
          }
        }
      }
    }
  }

  return false;
}

bool risc::Class::has_channel_pointers()
{
  SgDeclarationStatementPtrList &temp = get_ast_node()->get_members();

  for(std::vector<SgDeclarationStatement*>::iterator
      iter_decl  = temp.begin();
      iter_decl != temp.end();
      iter_decl++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter_decl);

    if (var_decl) {

      SgPointerType *type_pointer
        = isSgPointerType(
            var_decl->get_definition()->get_vardefn()->get_type());

      if (type_pointer) {

        // Simple pointer
        SgClassType *type = isSgClassType(type_pointer->get_base_type());

        if (type) {

          SgClassDeclaration *class_decl
            = isSgClassDeclaration(type->get_declaration());

          class_decl
            = isSgClassDeclaration(class_decl->get_definingDeclaration());

          PrimitiveChannel *prim_channel
            = risc::get_design()->find_prim_channel_definition(
                                    class_decl->get_definition());
          if (prim_channel) {
            return true;
          }


          HierarchicalChannel *hier_channel
            = risc::get_design()->find_channel_definition(
                                    class_decl->get_definition());
          if (hier_channel) {
            return true;
          }
        }


        // pointer to a pointer
        type_pointer = isSgPointerType(type_pointer->get_base_type());

        if (type_pointer) {

          SgClassType *class_type
            = isSgClassType(type_pointer->get_base_type());

          if (class_type) {

            SgClassDeclaration *class_decl
              = isSgClassDeclaration(class_type->get_declaration());

            class_decl
              = isSgClassDeclaration(class_decl->get_definingDeclaration());


            PrimitiveChannel *prim_channel
              = risc::get_design()->find_prim_channel_definition(
                                      class_decl->get_definition());
            if (prim_channel) {
              return true;
            }


            HierarchicalChannel *hier_channel
              = risc::get_design()->find_channel_definition(
                                      class_decl->get_definition());
            if (hier_channel) {
              return true;
            }
          }
        }
      }
    }
  }

  return false;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
