#ifndef PATH_INSTANCE_MAPPER_H_INCLUDED_
#define PATH_INSTANCE_MAPPER_H_INCLUDED_

#include "instance_tree.h"

#include <string>
#include <vector>

class DynamicModule;

namespace risc {

class PathInstanceMapper {

public:

  PathInstanceMapper();

  /**
   * \fn       print_mapping
   * \brief    The function prints the lookup table in the terminal
   */
  void print_mapping();

  typedef std::pair<InstanceTree, int> PathWithID;

  /**
   * \brief This is the lookup table from path to instance id
   */
  std::vector<PathWithID> path_to_id_mapping_;

  /**
   * \brief This variable counts what is the max of instances
   */
  int max_instances_;

  /**
   * \fn       perform_dynamic_analysis
   * \brief    This functions returns true if the path instance mapper should
   *           perform a dynamic analysis
   */
  virtual bool perform_dynamic_analysis() = 0;

  /**
   * \fn       get_instance_tree
   * \brief    This function returns for a given id and potential mapping
   *           candidates the corresponding instance tree.
   *           CAUTION: This function is only for the StaticPathInstanceMapper
   */
  virtual InstanceTree
  get_instance_tree(int id, std::list<InstanceTree> potential_instances);

  /**
   * \fn       get_id
   * \brief    This function returns for a given tree the corresponding instance
   *           id.
   *           CAUTION: This function is only for the StaticPathInstanceMapper
   */
  virtual int get_id(InstanceTree tree);

  /**
   * \fn       get_dynamic_tree_root
   * \brief    This function returns the root node of the dynamic analysis
   *           CAUTION: This function is only for the DynamicPathInstanceMapper
   */
  virtual DynamicModule* get_dynamic_tree_root();

  /**
   * \fn       get_global_variable_address()
   * \brief    This functions returns the address of the global variable with
   *           the name 'variable_name'
   *           CAUTION: This function is only for the DynamicPathInstanceMapper
   */
  virtual int* get_global_variable_address(std::string variable_name);

  /**
   * \fn       get_global_variable_name()
   * \brief    This functions translates the address of a global variable into
   *           the associated variable name
   *           CAUTION: This function is only for the DynamicPathInstanceMapper
   */
  virtual std::string get_global_variable_name(int *variable_address);
};

}; // end of namespace risc

#endif /* PATH_INSTANCE_MAPPER_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
