#include "class_is_derived_from.h"

#include <vector>

#include "rose_nodes.h"
#include "../internal_representation/design.h"
#include "../tools/systemc_definitions.h"

namespace risc {

namespace tools {

std::unordered_map<SgClassDefinition*,
                   std::unordered_map<SgClassDefinition*, bool> > lookup;

} // end namespace tools

} // end namespace risc

bool
risc::tools::is_derived_from(
    SgClassDefinition* derived_class,
    SgClassDefinition* base_class)
{
  std::unordered_map<SgClassDefinition*,
                     std::unordered_map<SgClassDefinition*, bool> >::iterator
    iter = lookup.find(base_class);


  // check if there the base class is already cached
  if(iter != lookup.end()) {

    std::unordered_map<SgClassDefinition*, bool>::iterator
      iter2 = iter->second.find(derived_class);

    // check if the derived class is already cached
    if(iter2 != iter->second.end()) {

      // return the cached result
      return iter2->second;

    } else {

      // No cached information available.
      // So we have to determine it and cache the result
      bool is_derived = is_derived_from_internal(derived_class, base_class);
      iter->second[derived_class] = is_derived;
      return is_derived;
    }

  } else {

    // Nothing is cached.
    // So, we have to determine the it and add the new unordered_map
    bool is_derived = is_derived_from_internal(derived_class, base_class);
    std::unordered_map<SgClassDefinition*, bool> temp;
    temp[derived_class] = is_derived;
    lookup[base_class] = temp;

    return is_derived;
  }
}

bool
risc::tools::is_derived_from_internal(
    SgClassDefinition* derived_class,
    SgClassDefinition* base_class)
{
  // FIXME This is a work around for the broken AST link of the base
  // class of sc_event_and_list
  if(base_class == NULL) {
    return false;
  }

  SgBaseClassPtrList base_classes = derived_class->get_inheritances();
  SgBaseClassPtrList::iterator iter = base_classes.begin();

  // iterate all direct base classes
  for(;iter != base_classes.end();iter++) {

    SgClassDeclaration *base_class_decl
      = isSgClassDeclaration(
          (*iter)->get_base_class()->get_definingDeclaration());

    // We check the class definition if they are the same.
    // We use the definition instead of the symbol for performance reasons
    if(base_class_decl) {

      if(base_class_decl->get_definition() == base_class) {
        return true;
      }
    }

    if(base_class_decl) {

      // recursive search
      if(is_derived_from(base_class_decl->get_definition(), base_class)) {
        return true;
      }

    } else {

      SgTemplateClassDeclaration *class_template_decl
        = isSgTemplateClassDeclaration((*iter)->get_base_class());

      if(class_template_decl) {

        // In this case we have a class template
        // So, we have to find the instances and check them.

        class_template_decl
          = isSgTemplateClassDeclaration(
              class_template_decl->get_firstNondefiningDeclaration());

        const std::vector<SgNode*> &template_instance_decl
          = RoseNodes::get_template_instance_decl();

        for(std::vector<SgNode*>::const_iterator
            iter  = template_instance_decl.begin();
            iter != template_instance_decl.end();
            iter++) {

          SgTemplateInstantiationDecl *template_class_decl_inst
            = isSgTemplateInstantiationDecl(*iter);

          SgTemplateClassDeclaration *temp
            = template_class_decl_inst->get_templateDeclaration();

          temp = isSgTemplateClassDeclaration(
                   temp->get_firstNondefiningDeclaration());

          if(class_template_decl == temp) {

            template_class_decl_inst
              = isSgTemplateInstantiationDecl(
                  template_class_decl_inst->get_definingDeclaration());

            // recursive search
            if(is_derived_from(template_class_decl_inst->get_definition(),
                  base_class)) {
              return true;
            }
          }
        }

      } else {

        if(!(*iter)->get_base_class()->get_definingDeclaration()) {

          // This can be sometimes quit annoying. So I turned it down.
          //std::cout << "Cannot find any definition of class: "
          //          << SageInterface::get_name((*iter)->get_base_class())
          //          << std::endl;
        }
      }
    }
  }

  return false;
}

bool
risc::tools::is_derived_from_sc_interface(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_interface());
}

bool
risc::tools::is_derived_from_sc_module(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_module())
         && !is_derived_from(derived_class, // because of typedef
             tools::SystemCDefinitions::get_sc_channel());
}

bool
risc::tools::is_derived_from_sc_port(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
              tools::SystemCDefinitions::get_sc_port());
}

bool
risc::tools::is_derived_from_sc_port_base(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_port_base());
}

bool
risc::tools::is_derived_from_sc_port_base(SgVarRefExp* var_ref)
{
  SgType *type = var_ref->get_type();
  SgReferenceType *ref_type = isSgReferenceType(type);
  if(ref_type) {
    type = ref_type->get_base_type();
  }

  SgClassType *class_type = isSgClassType(type);

  if(class_type == NULL) {
    return false;
  }

  SgClassDeclaration *class_decl
    = isSgClassDeclaration(class_type->get_declaration());

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  return is_derived_from(class_def,
             tools::SystemCDefinitions::get_sc_port_base());
}

bool
risc::tools::is_derived_from_sc_prim_channel(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_prim_channel());
}

bool
risc::tools::is_derived_from_sc_channel(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_channel());
}

bool
risc::tools::is_sc_in_port(SgTemplateInstantiationDefn* template_class)
{
  SgClassDeclaration *class_decl = template_class->get_declaration();
  SgTemplateInstantiationDecl *templ_class_decl =
    isSgTemplateInstantiationDecl(class_decl);

  return templ_class_decl->get_templateDeclaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_in()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_out_port(SgTemplateInstantiationDefn* template_class)
{
  SgClassDeclaration *class_decl = template_class->get_declaration();
  SgTemplateInstantiationDecl *templ_class_decl =
    isSgTemplateInstantiationDecl(class_decl);

  return templ_class_decl->get_templateDeclaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_out()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_inout_port(SgTemplateInstantiationDefn* template_class)
{
  SgClassDeclaration *class_decl = template_class->get_declaration();
  SgTemplateInstantiationDecl *templ_class_decl =
    isSgTemplateInstantiationDecl(class_decl);

  return templ_class_decl->get_templateDeclaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_inout()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_in_clk(SgClassDefinition *definition)
{
  return definition->get_declaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_in_clk()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_out_clk(SgClassDefinition *definition)
{
  return definition->get_declaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_out_clk()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_inout_clk(SgClassDefinition *definition)
{
  return definition->get_declaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_inout_clk()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_event(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_event())
    || (derived_class->get_declaration()->get_type() ==
         tools::SystemCDefinitions::get_sc_event()->
             get_declaration()->get_type());
}

bool
risc::tools::is_sc_event_and_list(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_event_and_list())
    || (derived_class->get_declaration()->get_type() ==
         tools::SystemCDefinitions::get_sc_event_and_list()->
             get_declaration()->get_type());
}

bool
risc::tools::is_sc_event_or_list(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_event_or_list())
    || (derived_class->get_declaration()->get_type() ==
         tools::SystemCDefinitions::get_sc_event_or_list()->
             get_declaration()->get_type());
}

// NEW: Identify types by properly utilizing the type system (i.e. stripType() and pointer comparison)
// NEW: and then use base classes of symbol/definition/variable/other for convenience functions!
// TODO: Finally, go over the entire code base and use these facilities (instead of local ad-hoc code).
// TODO: Only then we can get efficient and complete coverage! (RD, 07/14/18)

// I'll start with these below, one step at a time (RD, 07/14/18)
//
// This one can serve as documentation:
//
//bool
//risc::tools::is_sc_event(SgVarRefExp* var_ref)
//{
//  SgType *type = var_ref->get_type();                   // should be in convenience function
//  SgReferenceType *ref_type = isSgReferenceType(type);  // should be replaced by stripType()
//  if(ref_type) {                                        // (it'll fail for a 'const' object!)
//    type = ref_type->get_base_type();
//  }
//
//  return type ==                                        // this pointer comparison is good!
//    tools::SystemCDefinitions::get_sc_event()           // this is good too,
//                      ->get_declaration()->get_type();  // but we could use a convenience function:
//}                                                       // is_type_sc_event()
//
// The improved version follows:

bool
risc::tools::is_sc_event(SgVarRefExp* var_ref)  // (RD, 07/14/18)
{
  return is_type_sc_event(var_ref->get_type()
      ->stripType(SgType::STRIP_MODIFIER_TYPE
                  | SgType::STRIP_REFERENCE_TYPE
                  | SgType::STRIP_TYPEDEF_TYPE));
        // RD: for backward compatibility, we don't strip
        // SgType::STRIP_POINTER_TYPE | SgType::STRIP_ARRAY_TYPE
}

bool
risc::tools::is_sc_event_and_list(SgVarRefExp* var_ref) // (RD, 07/14/18)
{
  return is_type_sc_event_and_list(var_ref->get_type()
      ->stripType(SgType::STRIP_MODIFIER_TYPE
                  | SgType::STRIP_REFERENCE_TYPE
                  | SgType::STRIP_TYPEDEF_TYPE));
        // RD: for backward compatibility, we don't strip
        // SgType::STRIP_POINTER_TYPE | SgType::STRIP_ARRAY_TYPE
}

bool
risc::tools::is_sc_event_or_list(SgVarRefExp* var_ref)  // (RD, 07/14/18)
{
  return is_type_sc_event_or_list(var_ref->get_type()
      ->stripType(SgType::STRIP_MODIFIER_TYPE
                  | SgType::STRIP_REFERENCE_TYPE
                  | SgType::STRIP_TYPEDEF_TYPE));
        // RD: for backward compatibility, we don't strip
        // SgType::STRIP_POINTER_TYPE | SgType::STRIP_ARRAY_TYPE
}

// TODO: the 3 above functions actually don't belong here (nothing "derived" here),
// TODO: they should go into systemc_definitions.cpp

bool
risc::tools::is_sc_in_bool(SgClassDefinition* definition)
{
  SgTemplateInstantiationDefn *temp_inst_def =
    isSgTemplateInstantiationDefn(definition);

  if(temp_inst_def) {

    if(is_sc_in_port(temp_inst_def))  {

      SgTemplateInstantiationDecl *temp_inst_decl =
        isSgTemplateInstantiationDecl(temp_inst_def->get_declaration());

      if(temp_inst_decl->get_templateArguments().size() ==1 &&
          isSgTypeBool((temp_inst_decl->get_templateArguments().at(0))->
              get_type())) {
        return true;
      }
    }
  }

  return false;
}

bool
risc::tools::is_sc_fifo_in(SgVarRefExp* var_ref)
{
  SgType *type = var_ref->get_type();
  SgReferenceType *ref_type = isSgReferenceType(type);
  if(ref_type) {
    type = ref_type->get_base_type();
  }

  SgClassType *class_type = isSgClassType(type);

  if(class_type) {

    SgClassDeclaration *class_decl
      = isSgClassDeclaration(class_type->get_declaration());

    class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

    SgTemplateInstantiationDecl *template_instance_decl
      = isSgTemplateInstantiationDecl(class_decl);
    if(template_instance_decl) {

      SgTemplateClassDeclaration *template_decl
        = template_instance_decl->get_templateDeclaration();

      template_decl
        = isSgTemplateClassDeclaration(
            template_decl->get_definingDeclaration());

      return tools::SystemCDefinitions::get_sc_fifo_in()
                 == template_decl->get_definition();
    }
  }

  return false;
}

bool
risc::tools::is_sc_fifo_out(SgVarRefExp* var_ref)
{
  SgType *type = var_ref->get_type();
  SgReferenceType *ref_type = isSgReferenceType(type);
  if(ref_type) {
    type = ref_type->get_base_type();
  }

  SgClassType *class_type = isSgClassType(type);

  if(class_type) {

    SgClassDeclaration *class_decl
      = isSgClassDeclaration(class_type->get_declaration());

    class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

    SgTemplateInstantiationDecl *template_instance_decl
      = isSgTemplateInstantiationDecl(class_decl);
    if(template_instance_decl) {

      SgTemplateClassDeclaration *template_decl
        = template_instance_decl->get_templateDeclaration();

      template_decl
        = isSgTemplateClassDeclaration(
            template_decl->get_definingDeclaration());

      return tools::SystemCDefinitions::get_sc_fifo_out()
                 == template_decl->get_definition();
    }
  }

  return false;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
